# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021-2022 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests resource."""

from copy import deepcopy

from flask import g
from flask_resources import (
    from_conf,
    request_body_parser,
    request_parser,
    resource_requestctx,
    response_handler,
    route,
)
from invenio_records_resources.resources import RecordResource
from invenio_records_resources.resources.records.resource import (
    request_extra_args,
    request_headers,
)
from invenio_records_resources.resources.records.utils import search_preference

from ...customizations.event_types import CommentEventType


#
# Resource
#
class RequestCommentsResource(RecordResource):
    """Resource for Request comments for now."""

    list_view_args_parser = request_parser(
        from_conf("request_list_view_args"), location="view_args"
    )
    item_view_args_parser = request_parser(
        from_conf("request_item_view_args"), location="view_args"
    )
    search_args_parser = request_parser(
        from_conf("request_search_args"), location="args"
    )
    data_parser = request_body_parser(
        parsers=from_conf("request_body_parsers"),
        default_content_type=from_conf("default_content_type"),
    )

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        # Assignment of routes should be part of the
        # Config class
        routes = self.config.routes
        return [
            route("POST", routes["list"], self.create),
            route("POST", routes["reply"], self.reply),
            route("GET", routes["item"], self.read),
            route("PUT", routes["item"], self.update),
            route("DELETE", routes["item"], self.delete),
            route("GET", routes["timeline"], self.search),
            route("GET", routes["timeline_focused"], self.focused_list),
            route("GET", routes["replies"], self.get_replies),
            route("GET", routes["replies_focused"], self.focused_replies),
        ]

    @list_view_args_parser
    @request_extra_args
    @data_parser
    @response_handler()
    def create(self):
        """Create a top-level comment (no parent)."""
        data = deepcopy(resource_requestctx.data) if resource_requestctx.data else {}
        item = self.service.create(
            identity=g.identity,
            request_id=resource_requestctx.view_args["request_id"],
            data=data,
            event_type=CommentEventType,
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 201

    @item_view_args_parser
    @request_extra_args
    @data_parser
    @response_handler()
    def reply(self):
        """Create a reply to a comment."""
        data = deepcopy(resource_requestctx.data) if resource_requestctx.data else {}
        # Extract parent_id from the route (comment_id in the URL)
        parent_id = resource_requestctx.view_args["comment_id"]
        item = self.service.create(
            identity=g.identity,
            request_id=resource_requestctx.view_args["request_id"],
            parent_id=parent_id,
            data=data,
            event_type=CommentEventType,
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 201

    @item_view_args_parser
    @request_extra_args
    @response_handler()
    def read(self):
        """Read an event.

        Because each event has a unique id, we can disregard the request_id
        for now.
        """
        item = self.service.read(
            identity=g.identity,
            id_=resource_requestctx.view_args["comment_id"],
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @item_view_args_parser
    @request_extra_args
    @request_headers
    @data_parser
    @response_handler()
    def update(self):
        """Update a comment."""
        item = self.service.update(
            identity=g.identity,
            id_=resource_requestctx.view_args["comment_id"],
            data=resource_requestctx.data,
            revision_id=resource_requestctx.headers.get("if_match"),
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @item_view_args_parser
    @request_headers
    def delete(self):
        """Delete a comment."""
        self.service.delete(
            identity=g.identity,
            id_=resource_requestctx.view_args["comment_id"],
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return "", 204

    @list_view_args_parser
    @request_extra_args
    @search_args_parser
    @response_handler(many=True)
    def search(self):
        """Perform a search over EVENTS.

        Its primary purpose is as a batch read of events i.e. the timeline.
        """
        hits = self.service.search(
            identity=g.identity,
            request_id=resource_requestctx.view_args["request_id"],
            params=resource_requestctx.args,
            search_preference=search_preference(),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @list_view_args_parser
    @request_extra_args
    @search_args_parser
    @response_handler(many=True)
    def focused_list(self):
        """List the page containing the event with ID focus_event_id, or the first page of results if this is not found."""
        hits = self.service.focused_list(
            identity=g.identity,
            request_id=resource_requestctx.view_args["request_id"],
            focus_event_id=resource_requestctx.args.get("focus_event_id"),
            page_size=resource_requestctx.args.get("size"),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @item_view_args_parser
    @request_extra_args
    @search_args_parser
    @response_handler(many=True)
    def get_replies(self):
        """Get paginated replies for a specific comment."""
        hits = self.service.get_comment_replies(
            identity=g.identity,
            parent_id=resource_requestctx.view_args["comment_id"],
            params=resource_requestctx.args,
            search_preference=search_preference(),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @list_view_args_parser
    @request_extra_args
    @search_args_parser
    @response_handler(many=True)
    def focused_replies(self):
        """List the reply page containing the event with ID focus_event_id, or the first page of results if this is not found."""
        hits = self.service.focused_reply_list(
            identity=g.identity,
            parent_id=resource_requestctx.view_args["comment_id"],
            focus_reply_event_id=resource_requestctx.args.get("focus_event_id"),
            page_size=resource_requestctx.args.get("size"),
            expand=resource_requestctx.args.get("expand", False),
            params=resource_requestctx.args,
        )
        return hits.to_dict(), 200



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resolver and proxy for requests."""

from invenio_records_resources.references.entity_resolvers import (
    RecordPKProxy,
    RecordResolver,
)

from ..records.api import Request, RequestEvent
from ..services import RequestEventsServiceConfig, RequestsServiceConfig


class RequestResolver(RecordResolver):
    """Resolver for requests."""

    type_id = "request"

    def __init__(self):
        """Initialize the default record resolver."""
        super().__init__(
            record_cls=Request,
            service_id=RequestsServiceConfig.service_id,
            type_key=self.type_id,
            proxy_cls=RecordPKProxy,
        )

    def _reference_entity(self, entity):
        """Create a reference dict for the given request."""
        return {self.type_key: str(entity.id)}


class RequestEventResolver(RecordResolver):
    """Resolver for requests."""

    type_id = "request_event"

    def __init__(self):
        """Initialize the default record resolver."""
        super().__init__(
            record_cls=RequestEvent,
            service_id=RequestEventsServiceConfig.service_id,
            type_key=self.type_id,
            proxy_cls=RecordPKProxy,
        )

    def _reference_entity(self, entity):
        """Create a reference dict for the given request."""
        return {self.type_key: str(entity.id)}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dumps entity grant tokens into the indexed request record."""

from invenio_records.dumpers import SearchDumperExt
from invenio_records_resources.references import EntityGrant


class GrantTokensDumperExt(SearchDumperExt):
    """Grant tokens dumper.

    Responsible for serializing the required needs for accessing a request
    into the indexed request record.
    """

    grants_field = "grants"

    def __init__(self, *fields):
        """Constructor."""
        super().__init__()
        self.fields = fields

    def dump(self, request, data):
        """Dump grant tokens for entity fields into indexed request."""
        grants = []
        for field_name in self.fields:
            entity = getattr(request, field_name)
            if isinstance(entity, list):
                for e in entity:
                    for need in request.type.entity_needs(e):
                        grants.append(EntityGrant(field_name, need).token)
            else:
                for need in request.type.entity_needs(entity):
                    grants.append(EntityGrant(field_name, need).token)
        data[self.grants_field] = grants

    def load(self, data, request_cls):
        """Load the data."""
        data.pop(self.grants_field)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2025 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 - 2022 TU Wien.
# Copyright (C) 2025 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Event Schemas."""

from datetime import timezone

from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import (
    RAISE,
    fields,
)
from marshmallow_utils import fields as utils_fields

from invenio_requests.proxies import current_requests

from ..customizations.event_types import CommentEventType, EventType


class EventTypeMarshmallowField(fields.Str):
    """Serializes event type from `EventType` to string."""

    def _serialize(self, value, attr, data, **kwargs):
        if isinstance(value, EventType):
            value = value.type_id
        return value


class RequestEventSchema(BaseRecordSchema):
    """Base Event schema that other schemas should inherit from."""

    type = EventTypeMarshmallowField(dump_only=True)
    created_by = fields.Dict(dump_only=True)
    permissions = fields.Method("get_permissions", dump_only=True)
    parent_id = fields.String(allow_none=True, dump_only=True)
    children = fields.List(fields.Dict(), dump_only=True)

    def get_permissions(self, obj):
        """Return permissions to act on comments or empty dict."""
        service = current_requests.request_events_service
        request = self.context.get("request", None)
        permissions = {}
        is_comment = obj.type == CommentEventType

        if request is None:
            return {}

        if is_comment:
            permissions["can_update_comment"] = service.check_permission(
                self.context["identity"],
                "update_comment",
                event=obj,
                request=request,
            )
            permissions["can_delete_comment"] = service.check_permission(
                self.context["identity"],
                "delete_comment",
                event=obj,
                request=request,
            )
        else:
            # Other event types (e.g. log events) might be deleted comments, for which these permissions are inherently False.
            permissions["can_update_comment"] = False
            permissions["can_delete_comment"] = False

        permissions["can_reply_comment"] = service.check_permission(
            self.context["identity"],
            "reply_comment",
            event=obj,
            request=request,
        )

        return permissions


class RequestSchema(BaseRecordSchema):
    """Schema for requests.

    Note that the payload schema and the entity reference schemas (i.e. creator,
    receiver, and topic) are dynamically constructed and injected into this schema.
    """

    # load and dump
    type = fields.String()
    title = utils_fields.SanitizedUnicode(dump_default="")
    description = utils_fields.SanitizedUnicode()

    # Dump-only
    number = fields.String(dump_only=True)
    status = fields.String(dump_only=True)
    is_closed = fields.Boolean(dump_only=True)
    is_open = fields.Boolean(dump_only=True)
    expires_at = utils_fields.TZDateTime(
        timezone=timezone.utc, format="iso", dump_only=True
    )
    is_expired = fields.Boolean(dump_only=True)

    last_reply = fields.Nested(RequestEventSchema, dump_only=True)

    last_activity_at = utils_fields.TZDateTime(
        timezone=timezone.utc, format="iso", dump_only=True
    )

    class Meta:
        """Schema meta."""

        unknown = RAISE


class GenericRequestSchema(RequestSchema):
    """Generic request schema.

    CAUTION: This schema should not be used for the final validation of input
    data. Use the request type's own defined schema instead.

    This schema can be used in situations where you need to do basic validation
    or dumping of a request without the payload.

    This is used e.g. in Invenio-RDM-Records for dumping a request without
    having to know the specific request type.
    """

    created_by = fields.Dict()
    receiver = fields.Dict()
    topic = fields.Dict()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Base class for customizable actions on requests."""

from ..errors import NoSuchActionError
from ..proxies import current_events_service
from .event_types import LogEventType


class RequestAction:
    """Base class for actions on requests."""

    status_from = None
    """Required status of a request to run this action."""

    status_to = "created"
    """Status after execution of the action."""

    event_type = None
    """Defines an event which will be logged if defined."""

    log_event = True
    """Define if the action should be logged through a `LogEventType`."""

    def __init__(self, request):
        """Constructor."""
        self.request = request

    def can_execute(self):
        """Check whether the action can be executed.

        This checks if a given state transition for a request is allowed or
        not.

        :return: True if the action can be executed, False otherwise.
        """
        if self.status_from is None:
            return self.request.status is None
        else:
            return self.request.status in self.status_from

    def execute(self, identity, uow):
        """Execute the request action.

        :param identity: The identity of the executor.
        :param data: The passed input to the action.
        """
        self.request.status = self.status_to
        if self.log_event:
            event = LogEventType(payload=dict(event=self.status_to))
            _data = dict(payload=event.payload)
            current_events_service.create(
                identity, self.request.id, _data, event, uow=uow
            )


class RequestActions:
    """Namespace for RequestActions static calls."""

    @classmethod
    def get_action(cls, request, action_name):
        """Get the action registered under the given name.

        :param action_name: The registered name of the action.
        :return: The action registered under the given name.
        """
        try:
            return request.type.available_actions[action_name](request)
        except KeyError:
            raise NoSuchActionError(action=action_name)

    @classmethod
    def can_execute(cls, request, action_name):
        """Check whether an action can be executed.

        Perhaps data is sometimes useful for that check, so also included.
        """
        return cls.get_action(request, action_name).can_execute()

    @classmethod
    def execute(cls, identity, request, action_name, uow):
        """Have identity execute on request the action with the data."""
        return cls.get_action(request, action_name).execute(identity, uow)


#
# Specific actions
#
class CreateAction(RequestAction):
    """Create a request."""

    status_from = None
    status_to = "created"
    log_event = False


class CreateAndSubmitAction(RequestAction):
    """Create and submit a request."""

    status_from = None
    status_to = "submitted"
    log_event = False


class DeleteAction(RequestAction):
    """Delete a request."""

    status_from = ["created"]
    status_to = "deleted"


class SubmitAction(RequestAction):
    """Submit a request."""

    status_from = ["created"]
    status_to = "submitted"
    log_event = False


class AcceptAction(RequestAction):
    """Accept a request."""

    status_from = ["submitted"]
    status_to = "accepted"


class DeclineAction(RequestAction):
    """Decline a request."""

    status_from = ["submitted"]
    status_to = "declined"


class CancelAction(RequestAction):
    """Cancel a request."""

    status_from = ["submitted"]
    status_to = "cancelled"


class ExpireAction(RequestAction):
    """Expire a request."""

    status_from = ["submitted"]
    status_to = "expired"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

import inspect

from invenio_base.utils import entry_points

from . import config
from .registry import TypeRegistry
from .resources import (
    RequestCommentsResource,
    RequestCommentsResourceConfig,
    RequestsResource,
    RequestsResourceConfig,
)
from .services import (
    RequestEventsService,
    RequestEventsServiceConfig,
    RequestsService,
    RequestsServiceConfig,
    UserModerationRequestService,
)


class InvenioRequests:
    """Invenio-Requests extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        self.requests_service = None
        self.requests_resource = None
        self.request_comments_service = None
        self._schema_cache = {}
        self._events_schema_cache = {}
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.init_services(app)
        self.init_resources(app)
        self.init_registry(app)
        app.extensions["invenio-requests"] = self

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("REQUESTS_"):
                app.config.setdefault(k, getattr(config, k))

    def service_configs(self, app):
        """Customized service configs."""

        class ServiceConfigs:
            requests = RequestsServiceConfig.build(app)
            request_events = RequestEventsServiceConfig.build(app)

        return ServiceConfigs

    def init_services(self, app):
        """Initialize the service and resource for Requests."""
        service_configs = self.service_configs(app)

        self.requests_service = RequestsService(
            config=service_configs.requests,
        )
        self.request_events_service = RequestEventsService(
            config=service_configs.request_events,
        )
        self.user_moderation_requests_service = UserModerationRequestService(
            requests_service=self.requests_service,
        )

    def init_resources(self, app):
        """Init resources."""
        self.requests_resource = RequestsResource(
            service=self.requests_service,
            config=RequestsResourceConfig.build(app),
        )

        self.request_events_resource = RequestCommentsResource(
            service=self.request_events_service,
            config=RequestCommentsResourceConfig,
        )

    def init_registry(self, app):
        """Initialize the registry for Requests per type."""
        self.request_type_registry = TypeRegistry(
            app.config["REQUESTS_REGISTERED_TYPES"]
        )
        self.event_type_registry = TypeRegistry(
            app.config["REQUESTS_REGISTERED_EVENT_TYPES"]
        )
        self.entity_resolvers_registry = TypeRegistry(
            app.config["REQUESTS_ENTITY_RESOLVERS"]
        )
        # Load from entry points
        register_entry_point(
            self.request_type_registry, "invenio_requests.types", app=app
        )
        register_entry_point(self.request_type_registry, "invenio_requests.event_types")
        register_entry_point(
            self.entity_resolvers_registry, "invenio_requests.entity_resolvers"
        )


def register_entry_point(registry, ep_name, app=None):
    """Register types from an entry point."""
    for ep in entry_points(group=ep_name):
        loaded_ep = ep.load()
        type_cls = loaded_ep
        # Allow to load class from functions (note: classes are callable too)
        if app and inspect.isfunction(loaded_ep):
            type_cls = loaded_ep(app=app)
        registry.register_type(type_cls())


def finalize_app(app):
    """Finalize app.

    NOTE: replace former @record_once decorator
    """
    init(app)


def api_finalize_app(app):
    """Finalize app for api.

    NOTE: replace former @record_once decorator
    """
    init(app)


def init(app):
    """Register the module's services and indexers to the central registries."""
    svc_reg = app.extensions["invenio-records-resources"].registry
    idx_reg = app.extensions["invenio-indexer"].registry
    requests_ext = app.extensions["invenio-requests"]
    requests_service = requests_ext.requests_service
    events_service = requests_ext.request_events_service

    svc_reg.register(requests_service)
    svc_reg.register(events_service)

    idx_reg.register(requests_service.indexer, indexer_id="requests")
    idx_reg.register(events_service.indexer, indexer_id="events")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base class for creating custom event types of requests."""

import inspect

import marshmallow as ma
from marshmallow import RAISE, fields, validate
from marshmallow.validate import OneOf
from marshmallow_utils import fields as utils_fields

from ..proxies import current_requests


class EventType:
    """Base class for event types."""

    type_id = None
    """The unique and constant identifier for this type of event."""

    payload_schema = None
    """Schema for supported payload fields.

    Define it as a dictionary of fields mapping or callable that returns a dictionary:

    .. code-block:: python

        payload_schema = {
            "content": fields.String(),
            # ...
        }

        # or a callable
        def payload_schema():
            return {
                "content": fields.String(),
                # ...
            }
    """

    payload_required = False
    """Require the event payload."""

    allow_children = False
    """Allow this event type to have children (parent-child relationships).

    If True, events of this type can have a parent_id and children.
    If False, attempting to set a parent_id will raise an error.
    """

    def __init__(self, payload=None):
        """Constructor."""
        self.payload = payload or {}

    def __eq__(self, other):
        """Implement comparison test."""
        if inspect.isclass(other):
            # if a class was passed rather than an instance, try to instantiate it
            other = other()
        if isinstance(other, EventType):
            return self.type_id == other.type_id
        elif isinstance(other, str):
            return self.type_id == other
        raise Exception("unknown value")

    def __str__(self):
        """Return str(self)."""
        # Value used by marshmallow schemas to represent the type.
        return self.type_id

    def __repr__(self):
        """Return repr(self)."""
        return f"<RequestEventType '{self.type_id}'>"

    @classmethod
    def _create_marshmallow_schema(cls):
        """Create a marshmallow schema for this request type."""
        # Avoid circular imports
        from invenio_requests.services.schemas import RequestEventSchema

        additional_fields = {}

        # Raise on invalid payload keys
        class PayloadBaseSchema(ma.Schema):
            class Meta:
                unknown = RAISE

        # If a payload schema is defined, add it to the request schema
        if cls.payload_schema is not None:
            payload_schema = cls.payload_schema
            if callable(cls.payload_schema):
                payload_schema = payload_schema()

            payload_required = False
            if cls.payload_required is not None:
                payload_required = cls.payload_required
            additional_fields["payload"] = ma.fields.Nested(
                PayloadBaseSchema.from_dict(payload_schema), required=payload_required
            )

        # Dynamically create a schema from the fields defined
        # by the payload schema dict.
        return RequestEventSchema.from_dict(additional_fields)

    @classmethod
    def marshmallow_schema(cls):
        """Create a schema for the entire request including payload."""
        type_id = cls.type_id
        if type_id not in current_requests._events_schema_cache:
            current_requests._events_schema_cache[type_id] = (
                cls._create_marshmallow_schema()
            )
        return current_requests._events_schema_cache[type_id]


class LogEventType(EventType):
    """Log event type."""

    type_id = "L"

    def payload_schema():
        """Return payload schema as a dictionary."""
        # we need to import here because of circular imports
        from invenio_requests.records.api import RequestEventFormat

        return dict(
            event=fields.String(validate=validate.Length(min=1)),
            content=utils_fields.SanitizedHTML(validate=validate.Length(min=1)),
            format=fields.Str(
                validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
                load_default=RequestEventFormat.HTML.value,
            ),
        )


class ReviewersUpdatedType(EventType):
    """Reviewers updated event type."""

    type_id = "R"

    def payload_schema():
        """Return payload schema as a dictionary."""
        # we need to import here because of circular imports
        from invenio_requests.records.api import RequestEventFormat

        return dict(
            event=fields.String(validate=validate.Length(min=1)),
            content=utils_fields.SanitizedHTML(validate=validate.Length(min=1)),
            format=fields.Str(
                validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
                load_default=RequestEventFormat.HTML.value,
            ),
            reviewers=fields.List(
                fields.Dict(
                    keys=fields.String(validate=OneOf(("user", "group"))),
                    values=fields.String(required=True),
                )
            ),
        )


class CommentEventType(EventType):
    """Comment event type."""

    type_id = "C"

    allow_children = True
    """Comments support parent-child relationships (replies to comments)."""

    def payload_schema():
        """Return payload schema as a dictionary."""
        # we need to import here because of circular imports
        from invenio_requests.records.api import RequestEventFormat

        return dict(
            content=utils_fields.SanitizedHTML(
                required=True, validate=validate.Length(min=1)
            ),
            format=fields.Str(
                validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
                load_default=RequestEventFormat.HTML.value,
            ),
        )

    payload_required = True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""User moderation requests service errors."""

from invenio_i18n import lazy_gettext as _


class InvalidCreator(Exception):
    """Request creator is invalid."""

    description = _("Invalid creator for user moderation request.")


class OpenRequestAlreadyExists(Exception):
    """An open request already exists for the user."""

    description = _("There is already an open moderation request for this user.")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Search dumper for OpenSearch join relationships."""

from invenio_records.dumpers import SearchDumperExt


class ParentChildDumperExt(SearchDumperExt):
    """Search dumper extension for OpenSearch join relationships.

    This dumper sets the join relationship field for parent-child documents:
    - Parent events: {"name": "parent"}
    - Child events (replies): {"name": "child", "parent": parent_id}

    It also handles routing to ensure child documents are indexed on the
    same shard as their parent (required for join queries to work).
    """

    def dump(self, record, data):
        """Dump the join relationship data.

        Sets the parent_child field based on whether the record
        is a parent event or a child (reply).
        """
        if record.parent_id:
            # This is a child event (reply)
            data["parent_child"] = {"name": "child", "parent": str(record.parent_id)}
        else:
            # This is a parent event
            data["parent_child"] = {"name": "parent"}

    def load(self, data, record_cls):
        """Load the data.

        The join relationship is only used in the search index,
        not in the record data, so we remove it when loading.
        """
        data.pop("parent_child", None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request service results."""

from invenio_access.permissions import system_user_id
from invenio_records_resources.services.records.results import ExpandableField

from ..resolvers.registry import ResolverRegistry


class EntityResolverExpandableField(ExpandableField):
    """Expandable entity resolver field.

    It will use the Entity resolver registry to retrieve the service to
    use to fetch records and the fields to return when serializing
    the referenced record.
    """

    entity_proxy = None

    def ghost_record(self, value):
        """Return ghost representation of not resolved value."""
        return self.entity_proxy.ghost_record(value)

    def system_record(self):
        """Return the representation of a system user."""
        return self.entity_proxy.system_record()

    def get_value_service(self, value):
        """Return the value and the service via entity resolvers."""
        self.entity_proxy = ResolverRegistry.resolve_entity_proxy(value)
        v = self.entity_proxy._parse_ref_dict_id()
        _resolver = self.entity_proxy.get_resolver()
        service = _resolver.get_service()
        return v, service

    def pick(self, identity, resolved_rec):
        """Pick fields defined in the entity resolver."""
        return self.entity_proxy.pick_resolved_fields(identity, resolved_rec)


class MultiEntityResolverExpandableField(ExpandableField):
    """Expandable entity resolver field for multiple references.

    It uses the Entity resolver registry to retrieve the service to
    use to fetch records and the fields to return when serializing
    the referenced records.
    """

    def __init__(self, key):
        """Initialize the field."""
        super().__init__(key)
        self.entity_proxies = []

    def ghost_record(self, value):
        """Return ghost representation for unresolved values."""
        proxy = ResolverRegistry.resolve_entity_proxy(value)
        return proxy.ghost_record({"id": proxy._parse_ref_dict_id()})

    def system_record(self):
        """Return the representation of a system user."""
        return [
            ResolverRegistry.resolve_entity_proxy(value).system_record()
            for value in self.get_cached_value()
        ]

    def get_value_service(self, values):
        """Return a list of (value, service) tuples for multiple references."""
        self.entity_proxies = [
            ResolverRegistry.resolve_entity_proxy(value) for value in values
        ]

        results = []
        for proxy in self.entity_proxies:
            ref_id = proxy._parse_ref_dict_id()
            resolver = proxy.get_resolver()
            service = resolver.get_service()
            results.append((ref_id, service))
        return results

    def pick(self, identity, resolved_record):
        """Pick fields from resolved records based on the entity resolver."""
        for proxy in self.entity_proxies:
            if proxy._parse_ref_dict_id() == resolved_record["id"]:
                return proxy.pick_resolved_fields(identity, resolved_record)

    def add_dereferenced_record(self, service, value, resolved_rec):
        """Save the dereferenced record."""
        # mark the record as a "ghost" or "system" record i.e not resolvable
        if resolved_rec is None:
            if value == system_user_id:
                resolved_rec = self.system_record()
            else:
                resolver_type = self._get_resolver_type(service)
                resolved_rec = self.ghost_record({resolver_type: value})
        self._service_values[service][value] = resolved_rec

    def _get_resolver_type(self, service):
        """Return the resolver type."""
        for resolver in ResolverRegistry.get_registered_resolvers():
            if resolver._service_id == service.id:
                return resolver.type_id



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""User moderation request type."""

from .user_moderation import AcceptUserAction, DeclineUserAction, UserModerationRequest

__all__ = ("UserModerationRequest", "AcceptUserAction", "DeclineUserAction")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for managing referenced entities in request."""

from functools import partial

from invenio_records_resources.records.systemfields.entity_reference import (
    MultiReferenceEntityField,
    ReferencedEntityField,
    check_allowed_references,
)

from ...resolvers.registry import ResolverRegistry

EntityReferenceField = partial(
    ReferencedEntityField, resolver_registry=ResolverRegistry
)
"""An opinionated ReferenceEntityField with set ResolverRegistry."""

check_allowed_creators = partial(
    check_allowed_references,
    lambda r: r.type.creator_can_be_none,
    lambda r: r.type.allowed_creator_ref_types,
)
"""Check function specific for the ``created_by`` field of requests."""

check_allowed_receivers = partial(
    check_allowed_references,
    lambda r: r.type.receiver_can_be_none,
    lambda r: r.type.allowed_receiver_ref_types,
)
"""Check function specific for the ``receiver`` field of requests."""

check_allowed_topics = partial(
    check_allowed_references,
    lambda r: r.type.topic_can_be_none,
    lambda r: r.type.allowed_topic_ref_types,
)
"""Check function specific for the ``topic`` field of requests."""


check_allowed_reviewers = partial(
    check_allowed_references,
    lambda r: r.type.reviewers_can_be_none(),
    lambda r: r.type.allowed_reviewers_ref_types(),
)
"""Check function specific for the ``re`` field of requests."""


MultiEntityReferenceField = partial(
    MultiReferenceEntityField, resolver_registry=ResolverRegistry
)
"""An opinionated ReferenceEntityField with set ResolverRegistry."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Computed system fields for requests."""

from datetime import datetime

from invenio_db import db
from invenio_records_resources.records.systemfields.calculated import (
    CalculatedField,
)

from ...customizations import CommentEventType


class CachedCalculatedField(CalculatedField):
    """Cache-aware calculated field."""

    CACHE_MISS = object()

    def __init__(self, key=None, use_cache=True):
        """Constructor."""
        super().__init__(key=key, use_cache=use_cache)

    def calculate(self, record):
        """Fetch the calculated field from the cache if present."""
        # TODO: This logic should be pushed up to the `CalculatedField` class. If the
        # cache has a key for the system field, even if the value is `None`, it means
        # that the value was already calculated or explicitly cached, e.g. in
        # `post_load`.
        obj_cache = getattr(record, "_obj_cache", None)
        if obj_cache is not None and self.attr_name in obj_cache:
            return obj_cache[self.attr_name]
        return self.CACHE_MISS


class LastReply(CachedCalculatedField):
    """System field for getting the last reply event."""

    def __init__(self, key=None, use_cache=True):
        """Constructor."""
        super().__init__(key=key, use_cache=use_cache)

    def calculate(self, record):
        """Fetch the last reply event."""
        res = super().calculate(record)
        if res is not self.CACHE_MISS:
            return res

        RequestEvent = record.event_cls
        RequestEventModel = RequestEvent.model_cls

        last_comment = (
            db.session.query(RequestEventModel)
            .filter(
                RequestEventModel.request_id == record.id,
                RequestEventModel.type == CommentEventType.type_id,
            )
            .order_by(RequestEventModel.created.desc())
            .first()
        )

        if last_comment:
            return RequestEvent(data=last_comment.data, model=last_comment)

        return None

    def pre_dump(self, record, data, dumper=None):
        """Called before a record is dumped."""
        last_reply = getattr(record, self.attr_name)
        if last_reply:
            dumped_reply = last_reply.dumps()
            # Remove search-index-only fields that shouldn't be embedded
            dumped_reply.pop("parent_child", None)
            data[self.attr_name] = dumped_reply
        else:
            data[self.attr_name] = None

    def post_load(self, record, data, loader=None):
        """Called after a record was loaded."""
        RequestEvent = record.event_cls
        record.pop(self.attr_name, None)  # Remove the attribute from the record
        last_reply_dump = data.pop(self.attr_name, None)
        last_reply = None
        if last_reply_dump:
            last_reply = RequestEvent.loads(last_reply_dump)
        self._set_cache(record, last_reply)


class LastActivity(CachedCalculatedField):
    """System field for getting the last activity (derived from other fields)."""

    def __init__(self, key=None, use_cache=True):
        """Constructor."""
        super().__init__(key=key, use_cache=use_cache)

    def calculate(self, record):
        """Calculate the last activity."""
        res = super().calculate(record)
        if res is not self.CACHE_MISS:
            return res

        activity_dates = [record.model.updated]

        # Take into account the last comment if any
        # TODO: Extend this to other event types
        last_comment = record.last_reply
        if last_comment:
            activity_dates.append(last_comment.model.created)

        return max(activity_dates)

    def pre_dump(self, record, data, dumper=None):
        """Called before a record is dumped."""
        last_activity = getattr(record, self.attr_name)
        if last_activity:
            data[self.attr_name] = last_activity.isoformat()
        else:
            data[self.attr_name] = None

    def post_load(self, record, data, loader=None):
        """Called after a record was loaded."""
        record.pop(self.attr_name, None)  # Remove the attribute from the record
        last_activity_dump = data.pop(self.attr_name, None)
        last_activity = None
        if last_activity_dump:
            last_activity = datetime.fromisoformat(last_activity_dump)
        self._set_cache(record, last_activity)
